/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * safe.c -- versions of standard functions that do not return on
 * error
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>

#include "compat.h"
#include "format.h"
#include "options.h"
#include "safe.h"

#define SAFE_DIE(fmt, ...)					\
  do {								\
    fprintf(stderr, "%s:%d in function %s: " fmt ": %s\n",	\
	    file, line, func, __VA_ARGS__, strerror(errno));	\
    exit(-1);							\
  } while (0)

/* Like SAFE_DIE, but only takes one argument. This is required
   because it is not really possible to use __VA_ARGS__ portably when
   there are no arguments. */
#define SAFE_DIE_NOARGS(msg)					\
  do {								\
    fprintf(stderr, "%s:%d in function %s: " msg ": %s\n",	\
	    file, line, func, strerror(errno));			\
    exit(-1);							\
  } while (0)

/* Call another safe function from within one. */
#define SAFE_CALL(name, ...)				\
  _MU_SAFE_TR_NAME(name)(file, line, func, __VA_ARGS__)

void *_MU_SAFE_DECL(mu_xmalloc, size_t size) {
  void *ptr = malloc(size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xcalloc, size_t nmemb, size_t size) {
  void *ptr = calloc(nmemb, size);

  if (ptr)
    return ptr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xrealloc, void *ptr, size_t size) {
  void *res = realloc(ptr, size);

  if (res)
    return res;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to re-allocate %zu bytes of memory", size);
}

void *_MU_SAFE_DECL(mu_xreallocarray, void *ptr, size_t nmemb, size_t size) {
  void *res = reallocarray(ptr, nmemb, size);

  if (res)
    return res;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to re-allocate %zu elements, each of size %zu",
	   nmemb, size);
}

char *_MU_SAFE_DECL(mu_xstrdup, const char *str) {
  char *newstr = strdup(str);

  if (newstr)
    return newstr;

  /* If we're here, we failed to allocate memory. */
  SAFE_DIE("failed to allocate memory for a string of length %zu, \"%s\"",
	   strlen(str), str);
}

/*
 * These printf()-style functions return 'unsigned' simply to indicate
 * that the return value will never be negative. They will not,
 * however, return a value greater than INT_MAX.
 */

unsigned _MU_SAFE_DECL(mu_xasprintf, char **strp, const char *fmt, ...) {
  va_list ap;
  int ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(mu_xvasprintf, strp, fmt, ap);
  va_end(ap);

  return ret;
}

unsigned _MU_SAFE_DECL(mu_xvasprintf,
		       char **strp, const char *fmt, va_list ap) {
  int ret = vasprintf(strp, fmt, ap);
  if (ret < 0)
    SAFE_DIE("failed to allocate for a string with format \"%s\"", fmt);
  return ret;
}

/*
 * These are safe versions of functions provided by Mu itself.
 */

void _MU_SAFE_DECL(mu_xformat,
		   FILE *stream, unsigned short goal, unsigned short width,
		   unsigned short *col, unsigned short indent,
		   unsigned short subindent, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  SAFE_CALL(mu_xvformat,
	    stream, goal, width, col, indent, subindent, fmt, ap);
  va_end(ap);
}

char *_MU_SAFE_DECL(mu_xformat_string,
		    unsigned short goal, unsigned short width,
		    unsigned short *col, unsigned short indent,
		    unsigned short subindent, const char *fmt, ...) {
  va_list ap;
  char *ret;

  va_start(ap, fmt);
  ret = SAFE_CALL(mu_xvformat_string, goal, width,
		  col, indent, subindent, fmt, ap);
  va_end(ap);

  return ret;
}

void _MU_SAFE_DECL(mu_xvformat,
		   FILE *stream, unsigned short goal, unsigned short width,
		   unsigned short *col, unsigned short indent,
		   unsigned short subindent, const char *fmt, va_list ap) {
  if (mu_vformat(stream, goal, width, col, indent, subindent, fmt, ap))
    SAFE_DIE("failed to format a string with format \"%s\" to a file", fmt);
}

char *_MU_SAFE_DECL(mu_xvformat_string,
		    unsigned short goal, unsigned short width,
		    unsigned short *col, unsigned short indent,
		    unsigned short subindent, const char *fmt, va_list ap) {
  char *ret = mu_vformat_string(goal, width, col, indent, subindent, fmt, ap);
  if (!ret)
    SAFE_DIE("failed to format a string with format \"%s\" to a string", fmt);
  return ret;
}

void _MU_SAFE_DECL(mu_xformat_help,
		   FILE *stream, const char *prog_name,
		   const char *usage, const char *desc,
		   const char *notes, const MU_OPT *options) {
  if (mu_format_help(stream, prog_name, usage, desc, notes, options))
    SAFE_DIE_NOARGS("cannot print help message");
}

size_t _MU_SAFE_DECL(mu_xconcat_opts,
		     MU_OPT **res, const MU_OPT *a,
		     const MU_OPT *b, size_t extra) {
  size_t ret = mu_concat_opts(res, a, b, extra);
  if (!ret)
    SAFE_DIE("failed to concatenate two option lists: %p and %p", a, b);
  return ret;
}
